
#include <LiquidCrystal.h>
#include <EEPROM.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <HardwareSerial.h>
#include <PinChangeInterrupt.h>


LiquidCrystal lcd(11, 8, 7, 6, 5, 4);          // Configure pins required by lcd display

unsigned long speed_display_interval=1000000;  // interval for calculating the current speed, default 1s
unsigned long update_erg_braking=120000;       // interval for updating the braking resistance during erg mode, default to 0.12s
unsigned long ap2_update_interval=250000;      // interval for updating the ap2 ant+ module, default 0.25s
unsigned long display_update_interval=1000000; // interval for updating the display, default 1s

bool ticker_logo=false;

int sync_pulse=2;                              // pin for incoming 50Hz sync pulse from trainer
int pulse_count_pin=3;                         // pin for speed pulse input
int pwm_pin=9;                                 // pin for outputting pwm signal
int ap2_rts_pin=10;                            // pin used for rts/cts on AP2
int down_button=12;                            // pin used for the 'down' button
int up_button=13;                              // pin used for the 'up' button
int ok_button=19;
int cadence_pin=18;

bool up_pressed=false;
bool down_pressed=false;
bool ok_pressed=false;

volatile unsigned long speed_pulse_count=0;    // count of number of pulses received
volatile unsigned long pulse_time=1;           // this stores the time of the last pulse.
volatile unsigned long previous_pulse_time=0;  // time of the previous pulse received
unsigned long old_speed_calc_time=0;           // time of the previous speed calculation
float duration=0;                              // duration of the last group of pulses
String speed_units="mph ";                     // Just for output formatting to lcd

float brake_value=0;                           // initial value of the magnetic brake
int sync_pulse_start=640;                      // initial value for the start of the pwm pulse to control the brake
int sync_pulse_duration=101;                   // initial duration of the braking pulse
bool sync_done=false;                          // this is used so that the pwm output (change in brake value) is only reconfigured if a button has been pressed
int mains_frequency=50;                        // Mains AC frequency, this is important as the braking pulse is synchronised with this waveform.
float frequency_multiplier=1;                  // Default of 1 for 50Hz
int menu_option;

unsigned long display_update_time=0;           // time in milliseconds that the display was last updated

float speed_multiplier=1;                      // multiplier to convert between mph and kph, mph=1
float current_speed=0;                         // speed with current group of pulses read
float previous_speed_1=0;                      // previous speed
float previous_speed_2=0;                      // oldest previous speed calculated
float display_speed=0;                         // speed that will be displayed on the lcd
byte accumulated_distance=0;
byte workout_time=0;

unsigned char ant_read[254];                   // array to store the ant+ message bytes read in
byte rx_buffer_ptr=0;                          // pointer to the current position in the buffer

unsigned char ant_write[20];                   // array to store the ant+ message packet to be sent
unsigned long ap2_update_time=0;               // time of the last packet sent/read from/to the ap2 module
unsigned int send_manufacturer_counter=0;
bool manufacturer_info=false;

byte ant_event_counter=0;                      // incremented for each packet that is sent to the ap2 (as spec)
unsigned int current_power=0;                  // current calculated power
unsigned int accumulated_power=0;              // sum of power sent to the ap2, 16bit value that will rollover after 65535
float gradient_required=0;                     // gradient requested via ant
float power_required=0;                        // power level requested by ant
unsigned long erg_timer=0;
bool erg_power_control=false;                  // this is set true if a power request packet is received
bool gradient_control=false;                   // this is set if a gradient request packet is received

float power_calibration=7;                     // initial abitrary calibration value, should give 'ok' power figured, calibration should be run to update this

                                               // initial values for the power calibration curve
float ce1=4.14;                                // 4.129;     initial test power ~correct with calibration of 7.4
float ce2=1.37;                                // 1.365;
int power_map=1;

unsigned int memory_base=400;                  // base address to read/write stored values, this could be anywhere between 0 and just below 1k
char board_identifier[6]="tflow";
byte trainer_state=0b00100000;                 // initial trainer state of 'ready'
byte resistance_percentage=0x4D;               // brake value of 0 is 38.5% (0x4D is 38.5*2 in 0.5% increments) max resistance
String mode_symbol="-";
String mode_info="none";

void setup() 
{
                                                // Check if the 5 characters in flash base address match 'tflow' to check whether this arduino saved any previous settings
if(EEPROM.read(memory_base)==116 && EEPROM.read(memory_base+1)==102 && EEPROM.read(memory_base+2)==108 && EEPROM.read(memory_base+3)==111 && EEPROM.read(memory_base+4)==119)
  {
  EEPROM.get(memory_base+6, speed_multiplier);
  EEPROM.get(memory_base+10, mains_frequency);
  EEPROM.get(memory_base+12, frequency_multiplier);
  EEPROM.get(memory_base+16, power_calibration);
  EEPROM.get(memory_base+20, power_map);
  }
else                                            // if no settings found, save the defaults at the base address
  {
  EEPROM.put(memory_base, board_identifier);
  EEPROM.put(memory_base+6, speed_multiplier);
  EEPROM.put(memory_base+10, mains_frequency);
  EEPROM.put(memory_base+12, frequency_multiplier);
  EEPROM.put(memory_base+16, power_calibration);
  EEPROM.put(memory_base+20, power_map);
  }
if(speed_multiplier==1)
  speed_units="mph ";
else
  speed_units="kph ";

pinMode(pulse_count_pin, INPUT);               // configure 'pulse_count_pin' as an input

attachInterrupt(digitalPinToInterrupt(pulse_count_pin), speed_pulse, RISING);      // attach speed_pulse() interrupt routine to pin on the rising edge

pinMode(sync_pulse, INPUT);                    // configure 'sync_pulse' (2) as an input
attachInterrupt(digitalPinToInterrupt(sync_pulse), sync_input, FALLING);            // attach a sync_input interrupt routine to pin on the falling edge

pinMode(up_button, INPUT);                     // configure 'up_button' as an input
attachPCINT(digitalPinToPCINT(up_button), up_button_pressed, HIGH);
pinMode(down_button, INPUT);                   // configure 'down_button' as an input
attachPCINT(digitalPinToPCINT(down_button), down_button_pressed, HIGH);
pinMode(ok_button, INPUT);                     // configure 'ok_button' as an input
attachPCINT(digitalPinToPCINT(ok_button), ok_button_pressed, HIGH);

pinMode(cadence_pin, INPUT);

pinMode(pwm_pin, OUTPUT);

pinMode(ap2_rts_pin, INPUT);                   // configure 'ap2_rts_pin' as an input, to check that ap2 is ready to receive before sending a packet

                                               // set up Timer1 
  OCR1A = 100;                                 // initialize the comparator
  TIMSK1 = 0x03;                               // enable comparator A and overflow interrupts
  TCCR1A = 0x00;                               // timer control registers set for
  TCCR1B = 0x00;                               // normal operation, timer disabled.

  lcd.begin(16,4);                             // Initialise LCD

  Serial.begin(4800);                          // Configure serial port 0 to talk to the AP2 module, there is only 1 hardware serial port on the Arduino Nano.
  Serial.flush();
  delay(50);
  
  configure_ap2();                             // configure ap2 ant+ module

}

void loop() 
{
// ############################ Calculate Speed ############################       
  if ( (micros()-old_speed_calc_time)>speed_display_interval )   // check if at least 1 second have passed since last speed calculation
    calculate_speed();

// ############################# Check Buttons ############################# 
  if (up_pressed==true)
    {
    while (digitalRead(up_button)==HIGH);     // debounce following button press
    delay(200);
    up_pressed=false;
    increase_braking();
    gradient_control=false;                   // switch off gradient request mode if brake button is pressed
    if(erg_power_control==true)
      {
      erg_power_control=false;                // switch off erg mode if brake button pressed
      brake_value=0;                          // if switching off erg mode, power will probably drop to 0 and brake set to +8 to compensate, reset to 0.
      }
    trainer_state=0b00100000;                 // change 'trainer state' to 'ready'
    mode_symbol="-";
    mode_info="none";
    }

  if (down_pressed==true)
    {
    while (digitalRead(down_button)==HIGH);   // debounce following button press
    delay(200);
    down_pressed=false;
    decrease_braking();
    gradient_control=false;                   // switch off gradient request mode if brake button is pressed
    if(erg_power_control==true)
      {
      erg_power_control=false;                // switch off erg mode if brake button pressed
      brake_value=0;                          // if switching off erg mode, power will probably drop to 0 and brake set to +8 to compensate, reset to 0.
      }
    trainer_state=0b00100000;                 // change 'trainer state' to 'ready'
    mode_symbol="-";
    mode_info="none";
    }

  if (ok_pressed==true)
    {
    while (digitalRead(ok_button)==HIGH);      // debounce following button press
    delay(250);
    ok_pressed=false;
    configure_settings();
    }
    
// ######################## Updated the Ant+ Module ######################## 
  if ( (micros()-ap2_update_time)>ap2_update_interval )     // update ap2 every 0.25 seconds
    {
    calculate_current_power();
    send_new_ant_data();
    ap2_update_time=micros();
    }

// ####################### Process Ant+ Data Received ####################### 
  if (Serial.available()>0)                      // if there is any incoming data, get it and parse packet if complete
    get_received_data();
    
  if (erg_power_control==true && (micros()-erg_timer)>update_erg_braking)  // if a required power request is made, update braking to adjust power figure required
    {
    set_required_power();
    erg_timer=micros();
    }

// ############################# Update Display ############################# 
  if ( (micros()-display_update_time)>display_update_interval)   // update display every .5 second (default)
    {
    update_display();
    display_update_time=micros();
    }

  OCR1A = sync_pulse_start;                                // start of pulse, 1 for start of +ve, 156 for start of -ve ac
}

void get_received_data()
{
int count=0;
unsigned char checksum;

while (Serial.available() > 0)                             // read bytes in the rx queue to the ant_read[] array
  {
  ant_read[rx_buffer_ptr++]=Serial.read();

  if(rx_buffer_ptr==1 && ant_read[0]!=0xA4)                // if this is the start of a new packet and the first byte is not 0xA4 then discard it
    rx_buffer_ptr=0;

  if(rx_buffer_ptr>2 && ant_read[1]+4==rx_buffer_ptr)      // has a complete packet been read yet?
    {
    checksum=ant_read[0];                                  // calculate checksum and compare with value in packet, corrupted?
    for (count = 1; count < ant_read[1]+3; count++)
      checksum=checksum^ant_read[count];

    if(checksum==(unsigned char)ant_read[rx_buffer_ptr-1])                 // if checksum is ok then parse the packet
      {
      if(ant_read[2]==0x4F)
        {
        switch(ant_read[4])
          {
          case 0x30:                                                       // is this a basic resistance request packet?
            {
            resistance_percentage=(byte)(ant_read[11]);                    // percentage resistance requested 0-200, 0-100 in 0.5 increments
            brake_value=( (float)resistance_percentage/15.38 )-5;          // set brake_value from the percentage requested
            update_braking_signal();
            erg_power_control=false;
            gradient_control=true;
            erg_timer=0;
            trainer_state=0b00110000;                                      // change 'trainer state' to 'in use'
            mode_symbol="R";
            mode_info=String( ((float)resistance_percentage/2),1 )+String("%");
            break;
            }
          case 0x31:                                                       // is this a power request packet?
            {
            power_required=(float)((ant_read[11]*256)+ant_read[10])*0.25;  // calculate power from 2 byte value in packet
            erg_power_control=true;
            gradient_control=false;
            trainer_state=0b00110000;                                      // change 'trainer state' to 'in use'
            mode_symbol="P";
            mode_info=String( power_required,0 )+String("w");
            break;
            }
          case 0x33:                                                       // is this a gradient request packet? simulation mode
            {
            gradient_required=( (((float)ant_read[10]*256)+(float)ant_read[9])*0.01)-200; // calculate gradient from 2 byte value in packet
            brake_value=gradient_required*0.8;
            if(brake_value<-5)
              brake_value=-5;
            if(brake_value>8)
              brake_value=8;
            update_braking_signal();
            erg_power_control=false;
            gradient_control=true;
            trainer_state=0b00110000;                                      // change 'trainer state' to 'in use'
            resistance_percentage=(byte)( (brake_value+5)*15.38 );
            mode_symbol="G";
            mode_info=String( gradient_required,1 )+String("%");
            break;
            }
          case 0x46:                                                       // is this a status command request packet?
            {
            switch(ant_read[10])                                           // check which command has been sent and reply if needed
              {
              case 0x36:
                {
                ant_read[9]=ant_read[9]&127;                               // mask out 8th bit, first 7 bits of byte 9 are how many times to send requested packet
                for(count=0;count<ant_read[9];count++)
                  {
                  Send_FE_Capabilities();
                  delay(28);
                  }   
                break;           
                }             
              case 0x50:
                {
                ant_read[9]=ant_read[9]&127;                               // mask out 8th bit, first 7 bits of byte 9 are how many times to send requested packet
                for(count=0;count<ant_read[9];count++)
                  {
                  Send_Manufacturer_Info();
                  delay(28);
                  }
                break;
                }
              default:
                break;
              }
            break;
            }
          default:
            break;
          }
        }
        
      }
    rx_buffer_ptr=0;
    }
  }

}

void set_required_power()
{
  if (power_required<(current_power*0.99))
    decrease_braking();
  else
    if (power_required>(current_power*1.01))
      increase_braking();
  
}

void calculate_current_power()
{

switch(power_map)
  {
  case 1:
    { 
                               // this equation is really inefficient, but we have cycles to spare
    current_power=(unsigned int)((ce1+((brake_value)*0.128875))*pow(display_speed,(ce2+((brake_value)*0.017225))));     // calculate power
    current_power=(unsigned int)(current_power+(current_power*(7.45-power_calibration)/18));                            // adjust power on brake value
    break;
    }
  case 2:
    {
    current_power=(unsigned int)((8.6187*display_speed*1.60934)-49.384);
 //   current_power=(unsigned int)(current_power+(current_power*(7.45-power_calibration)/18));                            // adjust power on brake value
    break;
    }
  default:
    break;
  }

if(current_power<0)            // this should not happen but in some cases very close to 0 the power equation may give a -ve value just below 0
  current_power=0;
else
  if(current_power>9999)       // this should not happen, but just in case there is a glitch in the speed reading then this max value is set
    current_power=9999;
  
}

void sync_input()
{

  TCCR1B=0x04;                                 // start timer with divide by 1024 input, this should give just over 1/256 resolution of a 50Hz wave
}

ISR(TIMER1_COMPA_vect)
{                                              // comparator match
  digitalWrite(pwm_pin,HIGH);                  // set pwm_pin to high
  TCNT1 = 65536-sync_pulse_duration;           // trigger pulse width 0-590
}

ISR(TIMER1_OVF_vect)
{                                              // timer1 overflow
  digitalWrite(pwm_pin,LOW);                   // set pwm_pin to low
  TCCR1B = 0x00;                               // disable timer stopd unintended triggers
}


void calculate_speed()
{

  previous_speed_2=previous_speed_1;
  previous_speed_1=current_speed;

  duration=pulse_time-previous_pulse_time;
//  previous_pulse_time=pulse_time;                               // This is updated to match the current pulse time here, otherwise if no more pulses are received
                                                                // the duration never changes due to the interrupt not being called.
  if(duration==0)
    current_speed=0;                                            // If no pulses have been received then make speed=0, roller has not rotated 0.25 revolutions in the last second.
  else
    current_speed=(250000/duration)*0.212;
//    current_speed=(250000/duration)*0.203798736028;
  
  speed_pulse_count=0;                                          // reset pulse counter

  if ((pulse_time>4294367295) || (pulse_time<600000))           // micros() timer will flip back to 0 after ~71 minutes, this just makes sure
    current_speed=previous_speed_1;                             // there is no glitch in speed by using the old speed value once or twice during that period, less than ~1-2seconds

  if( previous_speed_1>(previous_speed_2*1.8) && (previous_speed_2<(current_speed*1.4)) )  // Check for missed pulse, if middle (of 3) pulse is greater than 1.8x first and first
    previous_speed_1=(previous_speed_2+current_speed)/2;                                   // is less than 1.4x oldest pulse, if so average the oldest and newest speed and assign to middle speed


  display_speed=(current_speed+previous_speed_1+previous_speed_2)/3;  // average the last 3 readings to smooth the output
//  display_speed=previous_speed_1;

  old_speed_calc_time=micros();                                       // record the time this of this reading

  lcd.setCursor(15,0);     // This bit does nothing apart from showing we haven't crashed (yet!)
  if(ticker_logo==true)
    {
    lcd.print("<");
    ticker_logo=false;
    }
  else
    {
    lcd.print(">");
    ticker_logo=true;
    }
}

void increase_braking()
{
  float working_brake_value;
    
    brake_value=(float)(round(brake_value*5)/5.0)+0.2;
      if (brake_value>8)                                                        // make sure max and min values of brake_value stay within -5 and +8 range
        brake_value=8;

    update_braking_signal();
}

void decrease_braking()
{
  float working_brake_value;

    brake_value=(float)(round(brake_value*5)/5.0)-0.2;
    if (brake_value<-5)                                                         // make sure max and min values of brake_value stay within -5 and +8 range
      brake_value=-5;

    update_braking_signal();
}

void update_braking_signal()
{
float working_brake_value;
    
    working_brake_value=brake_value+5;
    if (working_brake_value<8)
      {
      sync_pulse_start=640;
      working_brake_value=abs(working_brake_value-7.5);
      }   
    else
      {
      sync_pulse_start=10;
      working_brake_value=working_brake_value-7.5;
      }

    sync_pulse_duration=(int)(40*working_brake_value*frequency_multiplier+1);     // scale brake value to the first quarter wave cycle or third quarter for +ve braking
    sync_done=false;                                                              // set this flag so that the next sync interrupt can reconfigure the pwm output
  
}

void up_button_pressed()
{
    up_pressed=true;
}

void down_button_pressed()
{
    down_pressed=true;
}

void ok_button_pressed()
{
    ok_pressed=true;
}

void update_display()
{

  lcd.setCursor(0,0);
  lcd.print(String("Spd: ") + String((display_speed*speed_multiplier),1) + " " + speed_units );

  lcd.setCursor(0,1);
  lcd.print(String("Pwr: ") + String(current_power) + String(" ") );

  lcd.setCursor(0,2);
  lcd.print(String("Brk: ") + String(brake_value,1) + String(" ") );

  lcd.setCursor(0,3);
  lcd.print(String("Target: ") + mode_info + String("  ") );

  lcd.setCursor(15,3);
  lcd.print(mode_symbol);
}

void speed_pulse()                             // Interrupt routine to count incoming pulses from the speed sensor
{
    previous_pulse_time=pulse_time;
    speed_pulse_count++;
    pulse_time = micros();                     // save last pulse time, to allow for working out duration of pulses in speed calculation
}

void configure_settings()
{
bool settings_complete=false;

  menu_option=0;

  show_settings_page();
  display_settings_values();

  while(settings_complete==false)
    {
    lcd.setCursor(15,menu_option);
    lcd.blink();

  if (up_pressed==true)
    {
    while (digitalRead(up_button)==HIGH);     // debounce following button press
    delay(200);
    up_pressed=false;
    menu_option--;
    if (menu_option<0)
      {
      settings_complete=true;
      menu_option=0;
      }
    }

  if (digitalRead(down_button)==HIGH)
    {
    while (digitalRead(down_button)==HIGH);   // debounce following button press
    delay(200);
    down_pressed=false;
    menu_option++;
    if (menu_option>3)
      menu_option=3;
    }
 
    if (ok_pressed==true)
      {
      while (digitalRead(ok_button)==HIGH);   // debounce following button press
      delay(250);
      ok_pressed=false;
      
      switch(menu_option)                     // check which line the cursor is on when the 'ok' button was pressed
        {
        case 0:
          {
          calibrate_power();
          show_settings_page();
          break;
          }
        
        case 1:
          {
          power_map++;
          if(power_map>2)
            power_map=1;         
          EEPROM.put(memory_base+20, power_map);
          break;
          }

        case 2:
          {
          if(speed_multiplier==1)
            {
            speed_multiplier=1.60934;
            speed_units="kph ";
            }         
          else
            {
            speed_multiplier=1;
            speed_units="mph ";
            }

          EEPROM.put(memory_base+6, speed_multiplier);        // store new setting in eeprom
          break;
          }

        case 3:
          {
          if(mains_frequency==50)
            {
            mains_frequency=60;
            frequency_multiplier=0.83;  
            }
          else
            {
            mains_frequency=50;
            frequency_multiplier=1;
            }
            
          EEPROM.put(memory_base+10, mains_frequency);         // store new setting in eeprom
          EEPROM.put(memory_base+12, frequency_multiplier);
          break;
          }
        default:
          break;
        }
      display_settings_values();  
      }
    }

  lcd.noBlink();
  lcd.clear();
}

void show_settings_page()
{
//  lcd.clear();
//  lcd.setCursor(11,0);
//  lcd.print("Exit");

  lcd.setCursor(0,0);
  lcd.print("Calibrate       ");

  lcd.setCursor(0,1);
  lcd.print("Power Map: "+String(power_map)+"    ");

  lcd.setCursor(0,2);
  lcd.print("Speed:         ");

  lcd.setCursor(0,3);
  lcd.print("Frequency:      ");
}

void display_settings_values()
{
  lcd.setCursor(11,1);
  lcd.print(String(power_map));
  
  lcd.setCursor(12,2);
  if(speed_multiplier==1)
    lcd.print("mph");
  else
    lcd.print("kph");
   
  lcd.setCursor(11,3);
  if(mains_frequency==50)
    lcd.print("50Hz");
  else
    lcd.print("60Hz");
}

void calibrate_power()
{
bool exit_calibration=false;
volatile unsigned long rundown_time;
  
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("CALIBRATION");

  lcd.setCursor(0,1);
  lcd.print("Cycle over "+String(30*speed_multiplier,0) + speed_units);

  lcd.setCursor(0,3);
  lcd.print(String(power_calibration,2));

  delay(500);

  while( (current_speed<30) && (exit_calibration==false) )
    {

    calculate_speed();
    
    lcd.setCursor(0,2);
    lcd.print(String("Speed: ") + String((current_speed*speed_multiplier),1) + " " + speed_units );
    lcd.setCursor(14,3);
    lcd.print("X");
    if (ok_pressed==true)
      {
      while (digitalRead(ok_button)==HIGH);
      delay(250);
      ok_pressed=false;
      exit_calibration=true; 
      }
    delay(250);
    }
  if(exit_calibration==false)
    {
    lcd.setCursor(0,2);
    lcd.print("    STOP!!!     ");

    lcd.setCursor(0,3);
    lcd.print("Calculating...  ");

    while(display_speed>26)                                // wait for speed to start dropping and below 26mph to start timer, this should be more consistent.
      {
      calculate_speed();
      delay(100);
      }
      
    rundown_time=micros();                                 // record start time of rundown time
    while(display_speed>0.5)                               // wait until speed drops to 0.2mph to end the time count
      {
      calculate_speed();
      delay(100);
      }

    rundown_time=micros()-rundown_time;
    power_calibration=((float)rundown_time)/1000000;       // calibration number is rundown time in seconds

    lcd.setCursor(0,3);
    lcd.print(String(power_calibration,2) + String("               "));     // show the rundown time in seconds and wait for the 'ok' button to be pressed

    lcd.setCursor(10,3);
    lcd.print("Done");
    while (ok_pressed==false) delay(100);
    while (digitalRead(ok_button)==HIGH) delay(100);
    delay(200);
    ok_pressed=false;

    EEPROM.put(memory_base+16, power_calibration);          // store calibration value in eeprom to survive power cycles
    }
  delay(400);  
}

void configure_ap2()
{

  ResetAP2();
  delay(500);

  SetNetwork();
  delay(100);

  AssignCh();
  delay(100);

  SetChID();
  delay(100);

  SetFreq();
  delay(100);

  SetPeriod();
  delay(100);

  SetPower();
  delay(100);

  SetTimeout();
  delay(100);

  OpenChannel();
  delay(100);
  
}

void send_new_ant_data()
{
  ant_event_counter++;
  accumulated_power=accumulated_power+current_power;

  send_power();
  delay(28);                                      // These delays are not necessary but make diagnostics with the logic analyser easier see see packets.
  Send_General_FE_Data();
  delay(28);
  Send_General_FE_Data_DP17();
  delay(28);

  if(send_manufacturer_counter++%3==0)            // Send 2 DP80 & DP81 data packets alternatively every 4 data packets
    {
    if(manufacturer_info==false)
      {
      Send_Manufacturer_Info();
      delay(28);
      Send_Manufacturer_Info();
      delay(28);
      }
    else
      {
      Send_Manufacturer_Info2();
      delay(28);
      Send_Manufacturer_Info2();
      delay(28); 
      }
    manufacturer_info=!manufacturer_info;
    }

  if(send_manufacturer_counter==30)                // Send # packet every 30seconds, then reset counter.
     {
     send_manufacturer_counter=0;
     }

}

void send_ap2_packet(int write_packet_length)
{
int count;

while(digitalRead(ap2_rts_pin)==HIGH);             // wait for AP2 RTS pin to go low, ready to receive

for(count = 0 ; count < write_packet_length ; count++)
  Serial.write(ant_write[count]) ;
  
}

unsigned char calculate_ap2_checksum(int write_packet_length)
{
int count;

unsigned char checksum;

checksum=ant_write[0];
for (count = 1; count < write_packet_length; count++)
    checksum=checksum^ant_write[count];

return checksum;      
}

void ResetAP2()
{
  // Reset AP2
  ant_write[0] = 0xA4; // SYNC Byte
  ant_write[1] = 0x01; // LENGTH Byte
  ant_write[2] = 0x4A; // ID Byte
  ant_write[3] = 0x00;
  
  ant_write[4] = calculate_ap2_checksum(4);
  
  send_ap2_packet(5);
}

void SetNetwork()
{
    ant_write[0] = 0xA4;  // SYNC Byte
    ant_write[1] = 0x09;  // LENGTH Byte
    ant_write[2] = 0x46;  // ID Byte
    ant_write[3] = 0x00;  // Channel Number
    ant_write[4] = 0xB9;  // Data Byte
    ant_write[5] = 0xA5;  // Data Byte
    ant_write[6] = 0x21;  // Data Byte
    ant_write[7] = 0xFB;  // Data Byte
    ant_write[8] = 0xBD;  // Data Byte
    ant_write[9] = 0x72;  // Data Byte
    ant_write[10] = 0xC3; // Data Byte
    ant_write[11] = 0x45; // Data Byte
    
    ant_write[12] = calculate_ap2_checksum(12);
   
    send_ap2_packet(13);
}

void AssignCh()
{
   ant_write[0] = 0xA4; // SYNC Byte
   ant_write[1] = 0x03; // LENGTH Byte
   ant_write[2] = 0x42; // ID Byte
   ant_write[3] = 0x00; // Channel Number
   ant_write[4] = 0x10; // Channel Type
   ant_write[5] = 0x00; // Network ID
   
   ant_write[6] = calculate_ap2_checksum(6);
  
   send_ap2_packet(7);
}

void SetChID()
{
    ant_write[0] = 0xA4; // SYNC Byte
    ant_write[1] = 0x05; // LENGTH Byte
    ant_write[2] = 0x51; // Assign Channel ID - 0x51
    ant_write[3] = 0x00; // channel number
    ant_write[4] = 0xCF; // Device number
    ant_write[5] = 0x00; // Device number? Pairing request 0 - OFF
    ant_write[6] = 0x11; // Device type ID  0x11 - FEC      0x0B - Power Meter
    ant_write[7] = 0x05; // Transmission type
    
    ant_write[8] = calculate_ap2_checksum(8);
    
    send_ap2_packet(9);
}

void SetFreq()
{
    ant_write[0] = 0xA4; // SYNC Byte
    ant_write[1] = 0x02; // LENGTH Byte
    ant_write[2] = 0x45; // Set Channel RF Freq 0x45
    ant_write[3] = 0x00; // Channel number
    ant_write[4] = 0x39; // Frequency used for FE-C devices
    
    ant_write[5] = calculate_ap2_checksum(5);
   
    send_ap2_packet(6);  
}

void SetPeriod()
{
    ant_write[0] = 0xA4; // SYNC Byte
    ant_write[1] = 0x03; // LENGTH Byte
    ant_write[2] = 0x43; // Set channel period 0x43
    ant_write[3] = 0x00; // Channel number
    ant_write[4] = 0xF6; // Messaging Period byte1
    ant_write[5] = 0x1F; // Messaging period byte2
    
    ant_write[6] = calculate_ap2_checksum(6);
   
    send_ap2_packet(7);  
}

void SetPower()
{
    ant_write[0] = 0xA4; // SYNC Byte
    ant_write[1] = 0x02; // LENGTH Byte
    ant_write[2] = 0x47; // Set Tx Power - 0x47
    ant_write[3] = 0x00; // Channel Number
    ant_write[4] = 0x03; // Tx power - 0dB (maximum)
    
    ant_write[5] = calculate_ap2_checksum(5);
   
    send_ap2_packet(6);  
}

void SetTimeout()
{
    ant_write[0] = 0xA4; // SYNC Byte
    ant_write[1] = 0x02; // LENGTH Byte
    ant_write[2] = 0x44; // Set Channel Search Timeout 0x44
    ant_write[3] = 0x00; // Channel number
    ant_write[4] = 0x1E; // Set timeout
    
    ant_write[5] = calculate_ap2_checksum(5);
       
    send_ap2_packet(6);  
}

void OpenChannel()
{
    ant_write[0] = 0xA4; // SYNC Byte
    ant_write[1] = 0x01; // LENGTH Byte
    ant_write[2] = 0x4B; // ID Byte 0x4B
    ant_write[3] = 0x00;
    
    ant_write[4] = calculate_ap2_checksum(4);
   
    send_ap2_packet(5);  
}

void send_power()
{
    ant_write[0] = 0xA4;                                  // SYNC Byte
    ant_write[1] = 0x09;                                  // LENGTH Byte
    ant_write[2] = 0x4E;                                  // 0x4E
    ant_write[3] = 0x00;                                  // Channel number
    ant_write[4] = 0x19;                                  // fec power/cadence page identifier
    ant_write[5] = ant_event_counter;                     // Event count
    ant_write[6] = 0xFF;                                  // Instant Cadence (invalid 0xFF)
    ant_write[7] = lowByte(accumulated_power);            // Accumulated power LSB
    ant_write[8] = highByte(accumulated_power);           // Accumulated power MSB
    ant_write[9] = lowByte(current_power);                // Instant power LSB
    ant_write[10] = byte(0x7F & highByte(current_power)); // Instant power MSN
    ant_write[11] = trainer_state;
    
    ant_write[12] = calculate_ap2_checksum(12);
   
    send_ap2_packet(13);  
}

void Send_Manufacturer_Info()
{
    ant_write[0] = 0xA4;                                  // SYNC Byte
    ant_write[1] = 0x09;                                  // LENGTH Byte
    ant_write[2] = 0x4E;                                  // 0x4E
    ant_write[3] = 0x00;                                  // Channel number
    ant_write[4] = 0x50;                                  // Manufacturer’s Information identifier
    ant_write[5] = 0xFF;                                  // Reserved
    ant_write[6] = 0xFF;                                  // Reserved
    ant_write[7] = 0x01;                                  // HW Revision
    ant_write[8] = 0xFF;                                  // Manufacturer ID LSB 0xFF - Development
    ant_write[9] = 0x00;                                  // Manufacturer ID MSB
    ant_write[10] = 0x01;                                 // Model Number LSB
    ant_write[11] = 0x00;                                 // Model Number MSB
    
    ant_write[12] = calculate_ap2_checksum(12);
   
    send_ap2_packet(13);  
}

void Send_Manufacturer_Info2()
{
    ant_write[0] = 0xA4;                                  // SYNC Byte
    ant_write[1] = 0x09;                                  // LENGTH Byte
    ant_write[2] = 0x4E;                                  // 0x4E
    ant_write[3] = 0x00;                                  // Channel number
    ant_write[4] = 0x51;                                  // Manufacturer’s Information identifier (81)
    ant_write[5] = 0xFF;                                  // Reserved
    ant_write[6] = 0xFF;                                  // Reserved
    ant_write[7] = 0x01;                                  // SW Revision
    ant_write[8] = 0xFF;                                  // Serial Number bits 0-7
    ant_write[9] = 0xFF;                                  // Serial Number bits 8-15
    ant_write[10] = 0xFF;                                 // Serial Number bits 16-31
    ant_write[11] = 0xFF;                                 // Serial Number bits 32-47
    
    ant_write[12] = calculate_ap2_checksum(12);
   
    send_ap2_packet(13);  
}

void Send_General_FE_Data()
{
int instant_speed;

instant_speed=(int)(current_speed*0.44704);               // ~current speed in m/s
accumulated_distance=accumulated_distance+(byte)instant_speed; // ~ accumulated distance travelled in m, rolls over after 256m.
workout_time++;                                           // ~duration of workout, as this packer is sent ~ every 1 second, add 1 each time.
 
    ant_write[0] = 0xA4;                                  // SYNC Byte
    ant_write[1] = 0x09;                                  // LENGTH Byte
    ant_write[2] = 0x4E;                                  // 0x4E
    ant_write[3] = 0x00;                                  // Channel number
    ant_write[4] = 0x10;                                  // General FE Data
    ant_write[5] = 0x19;                                  // Equipment Type - 0x19 Trainer/Stationary Bike
    ant_write[6] = workout_time;                          // Accumulated value of the elapsed time since start of workout
    ant_write[7] = accumulated_distance;                  // Accumulated value of the distance traveled since start of workout
    ant_write[8] = lowByte(instant_speed);                // Instantaneous speed in m/s LSB
    ant_write[9] = highByte(instant_speed);               // Instantaneous speed in m/s MSB
    ant_write[10] = 0xFF;                                 // Instantaneous heart rate, 0xFF indicates invalid
    ant_write[11] = trainer_state;                        // Bits 0-3 Capabilities Bit Field, bits 4-7 FE State Bit Field
    
    ant_write[12] = calculate_ap2_checksum(12);
    
    send_ap2_packet(13);  
}

void Send_General_FE_Data_DP17()
{
    ant_write[0] = 0xA4;                                  // SYNC Byte
    ant_write[1] = 0x09;                                  // LENGTH Byte
    ant_write[2] = 0x4E;                                  // 0x4E
    ant_write[3] = 0x00;                                  // Channel number
    ant_write[4] = 0x11;                                  // General Settings data page (17 0x11)
    ant_write[5] = 0xFF;                                  // Reserved
    ant_write[6] = 0xFF;                                  // Reserved
    ant_write[7] = 0xFF;                                  // Cycle Length - set to invalid 0xFF
    ant_write[8] = 0xFF;                                  // Incline LSB, 0xFF invalid
    ant_write[9] = 0x7F;                                  // Incline MSB, 0x7F invalid
    ant_write[10] = resistance_percentage;                // Percentage of max resistance in 0.5% increments, brake levels -5 to +8, 0 is 38.5%, should be flat road 'feel'
    ant_write[11] = trainer_state;                        // Capabilities Bit Field
    
    ant_write[12] = calculate_ap2_checksum(12);
    
    send_ap2_packet(13);  
}

void Send_FE_Capabilities()
{
    ant_write[0] = 0xA4;                                  // SYNC Byte
    ant_write[1] = 0x09;                                  // LENGTH Byte
    ant_write[2] = 0x4E;                                  // 0x4E
    ant_write[3] = 0x00;                                  // Channel number
    ant_write[4] = 0x36;                                  // FE Capabilities
    ant_write[5] = 0xFF;                                  // Reserved
    ant_write[6] = 0xFF;                                  // Reserved
    ant_write[7] = 0xFF;                                  // Reserved
    ant_write[8] = 0xFF;                                  // Reserved
    ant_write[9] = 0x64;                                  // Maximum resistance (N) of trainer LSB, 0xff invalid. Set to 100Nm for Tacx Flow, a guess!
    ant_write[10] = 0x00;                                 // Maximum resistance (N) of trainer MSB,  0xff invalid
    ant_write[11] = 0b00000011;                           // Capabilities Bit Field
    
    ant_write[12] = calculate_ap2_checksum(12);
    
    send_ap2_packet(13);  
}
